/*
** File: admxrc2_mmap.c  
** Project: ADMXRC2 API module library
** Purpose: Implements VxWorks-specific functions for mapping memory in
**          ADMXRC2 API module library.
**
** (C) Copyright Alpha Data 2016
*/

#include <ioLib.h>

#include "admxrc2_platform.h"
#include <ioctl_dfvxw.h>

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
# define HAS_VX_IOCTL_ARG_T (1)
#else
# define HAS_VX_IOCTL_ARG_T (0)
#endif

#define TRACE 0

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

struct _MappedRegion;
typedef struct _MappedRegion MappedRegion;

struct _MappedRegion {
  MappedRegion* pPrev;
  MappedRegion* pNext;
  struct {
    ADMXRC2_HANDLE hDevice;
    unsigned int regionTag;
    uint64_t offset;
    uint64_t length;
  } parameters;
  struct {
    void* pVirtual;
  } mapped;
};

static struct _MappedList {
  MappedRegion* pHead;
  MappedRegion* pTail;
} g_mappedList;
static SEM_ID g_mappedListLock = NULL;
static int g_bInitialized = FALSE;

void
admxrc2MmapInitialize(
  void)
{
  if (!g_bInitialized) {
#if TRACE
    fprintf(stderr, "initialize: initializing globals\n");
#endif
    g_mappedListLock = semBCreate(SEM_Q_PRIORITY, SEM_FULL);
    g_mappedList.pHead = NULL;
    g_mappedList.pTail = NULL;
    g_bInitialized = TRUE;
  }
}

static void
addRegion(
  MappedRegion* pMappedRegion)
{
  semTake(g_mappedListLock, WAIT_FOREVER);
  pMappedRegion->pPrev = g_mappedList.pTail;
  pMappedRegion->pNext = NULL;
  if (NULL == g_mappedList.pHead) {
    g_mappedList.pHead = pMappedRegion;
    g_mappedList.pTail = pMappedRegion;
  } else {
    g_mappedList.pTail->pNext = pMappedRegion;
    g_mappedList.pTail = pMappedRegion;
  }
  semGive(g_mappedListLock);
}

/* Must be called while holding the semaphore */
static void
removeRegion(
  MappedRegion* pMappedRegion)
{
  if (NULL == pMappedRegion->pPrev) {
    g_mappedList.pHead = pMappedRegion->pNext;
  } else {
    pMappedRegion->pPrev->pNext = pMappedRegion->pNext;
  }
  if (NULL == pMappedRegion->pNext) {
    g_mappedList.pTail = pMappedRegion->pPrev;
  } else {
    pMappedRegion->pNext->pPrev = pMappedRegion->pPrev;
  }
  pMappedRegion->pPrev = NULL;
  pMappedRegion->pNext = NULL;
}

static MappedRegion*
findAndRemoveRegion(
  ADMXRC2_HANDLE hDevice,
  const void* pVirtual,
  bool_t* pbFoundHandle)
{
  MappedRegion* p;
  bool_t bFoundHandle = FALSE;

  semTake(g_mappedListLock, WAIT_FOREVER);
  p = g_mappedList.pHead;
  while (NULL != p) {
    if (p->parameters.hDevice == hDevice) {
      bFoundHandle = TRUE;
      if (p->mapped.pVirtual == pVirtual) {
        break;
      }
    }
    p = p->pNext;
  }
  if (NULL != p) {
#if TRACE
    fprintf(stderr, "findAndRemoveRegion: removing region, hDevice=%d pVirtual=%p pRegion=%p\n", hDevice, pVirtual, p);
#endif
    removeRegion(p);
  } else {
#if TRACE
    fprintf(stderr, "findAndRemoveRegion: failed to find region, hDevice=%d pVirtual=%p pRegion=%p\n", hDevice, pVirtual, p);
#endif
  }
  semGive(g_mappedListLock);

  *pbFoundHandle = bFoundHandle;
  return p;
}

/* Must be called while holding mutex */
static MappedRegion*
findRegion(
  ADMXRC2_HANDLE hDevice,
  unsigned int regionTag,
  unsigned int ordinal,
  bool_t* pbFoundHandle)
{
  MappedRegion* p;
  bool_t bFoundHandle = FALSE;

  p = g_mappedList.pHead;
  while (NULL != p) {
    if (p->parameters.hDevice == hDevice) {
      bFoundHandle = TRUE;
      if (p->parameters.regionTag == regionTag) {
        if (ordinal == 0) {
          break;
        }
        ordinal--;
      }
    }
    p = p->pNext;
  }

  *pbFoundHandle = bFoundHandle;
  return p;
}

ADMXRC2_STATUS
admxrc2MapRegion(
  ADMXRC2_HANDLE hAdmxrc2,
  unsigned int regionTag,
  uint64_t offset,
  uint64_t length,
  void** ppVirtualBase)
{
  DfVxwIoctlMmap ios;
  MappedRegion* pMappedRegion;
  int err;

  admxrc2MmapInitialize();
  
  /* Sanity checks */
  if (0 == length || offset + length < offset) {
    return ADMXRC2_INVALID_REGION;
  }

  pMappedRegion = (MappedRegion*)malloc(sizeof(MappedRegion));
  if (NULL == pMappedRegion) {
    return ADMXRC2_NO_MEMORY;
  }

  ios.in.regionTag = regionTag;
  ios.in.offset = offset;
  ios.in.length = length;
#if HAS_VX_IOCTL_ARG_T
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MMAP, sizeof(ios.in), sizeof(ios.out)), (_Vx_ioctl_arg_t)&ios);
#else
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MMAP, sizeof(ios.in), sizeof(ios.out)), (int)&ios);
#endif
  if (OK == err) {
    pMappedRegion->parameters.hDevice = hAdmxrc2;
    pMappedRegion->parameters.regionTag = regionTag;
    pMappedRegion->parameters.offset = offset;
    pMappedRegion->parameters.length = length;
    pMappedRegion->mapped.pVirtual  = ios.out.pVirtualBase;
  #if TRACE
    fprintf(stderr, "admxrc2MapRegion: adding region, hDevice=%d pRegion%p\n", hAdmxrc2, pMappedRegion);
  #endif
    addRegion(pMappedRegion);
    
    *ppVirtualBase = ios.out.pVirtualBase;
    return ADMXRC2_SUCCESS;
  } else {
    free(pMappedRegion);
    if (ERROR == err) {
      switch (errno) {
      case EBADF:
        return ADMXRC2_INVALID_HANDLE;

      case EFAULT:
      case ENOSYS:
      default:
        fprintf(stderr, "admxrc2MapRegion: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC2_UNKNOWN_ERROR;
      }
    } else {
      switch ((DfVxwMmapStatus)err) {
      case DfVxwMmapInvalidRegion:
        return ADMXRC2_INVALID_REGION;

      case DfVxwMmapNoMemory:
        return ADMXRC2_NO_MEMORY;

      case DfVxwMmapRegionTooLarge:
        return ADMXRC2_REGION_TOO_LARGE;

      case DfVxwMmapGeneralFailure:
      default:
        return ADMXRC2_UNKNOWN_ERROR;
      }
    }
  }
}

ADMXRC2_STATUS
admxrc2UnmapRegion(
  ADMXRC2_HANDLE hAdmxrc2,
  void* pVirtualBase)
{
  DfVxwIoctlMunmap ios;
  MappedRegion* p;
  bool_t bFoundHandle;
  int err;

  admxrc2MmapInitialize();
  
  p = findAndRemoveRegion(hAdmxrc2, pVirtualBase, &bFoundHandle);
  if (NULL == p) {
    return bFoundHandle ? ADMXRC2_INVALID_REGION : ADMXRC2_INVALID_HANDLE;
  }
  free(p);
  
  ios.in.pVirtualBase = (void*)pVirtualBase;
#if HAS_VX_IOCTL_ARG_T
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MUNMAP, sizeof(ios.in), 0), (_Vx_ioctl_arg_t)&ios);
#else
  err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MUNMAP, sizeof(ios.in), 0), (int)&ios);
#endif
  if (OK == err) {
    return ADMXRC2_SUCCESS;
  } else {
    if (ERROR == err) {
      switch (errno) {
      case EBADF:
        return ADMXRC2_INVALID_HANDLE;

      case EFAULT:
        return ADMXRC2_UNKNOWN_ERROR;

      case ENOSYS:
      default:
        fprintf(stderr, "admxrc2UnmapRegion: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC2_UNKNOWN_ERROR;
      }
    } else {
      switch (err) {
      case DfVxwMunmapInvalidRegion:
        return ADMXRC2_INVALID_REGION;

      case DfVxwMunmapGeneralFailure:
      default:
        return ADMXRC2_UNKNOWN_ERROR;
      }
    }
  }
}

void
admxrc2UnmapAllRegions(
  ADMXRC2_HANDLE hAdmxrc2)
{
  DfVxwIoctlMunmap ios;
  MappedRegion* p, * pNext;
  int err;
  
  admxrc2MmapInitialize();
  
  semTake(g_mappedListLock, WAIT_FOREVER);
  p = g_mappedList.pHead;
  while (NULL != p) {
    pNext = p->pNext;
    if (p->parameters.hDevice == hAdmxrc2) {
#if TRACE
      fprintf(stderr, "admxrc2UnmapAllRegions: removing region, hDevice=%d pRegion=%p\n", hAdmxrc2, p);
#endif
      removeRegion(p);
#if TRACE
      fprintf(stderr, "admxrc2UnmapAllRegions: calling munmap, pVirtual=%p\n",
        p->mapped.pVirtual);
#endif
      ios.in.pVirtualBase = (void*)p->mapped.pVirtual;
      free(p);
    #if HAS_VX_IOCTL_ARG_T
      err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MUNMAP, sizeof(ios.in), 0), (_Vx_ioctl_arg_t)&ios);
    #else
      err = ioctl(hAdmxrc2, DFVXW_MAKE_IOCTL(DFVXW_IOCTLCODE_MUNMAP, sizeof(ios.in), 0), (int)&ios);
    #endif
      if (OK != err) {
        fprintf(stderr, "admxrc2UnmapAllRegions: munmap() failed, pVirtual=%p err=%d\n",
          p->mapped.pVirtual, err);
      }
    }
    p = pNext;
  }
  semGive(g_mappedListLock);
}

ADMXRC2_STATUS
admxrc2FindRegion(
  ADMXRC2_HANDLE hAdmxrc2,
  unsigned int regionTag,
  unsigned int ordinal,
  void** ppVirtualBase,
  uint64_t* pOffset,
  uint64_t* pLength)
{
  MappedRegion* pMappedRegion;
  bool_t bFoundHandle = FALSE;

  admxrc2MmapInitialize();
  
  semTake(g_mappedListLock, WAIT_FOREVER);
  pMappedRegion = findRegion(hAdmxrc2, regionTag, ordinal, &bFoundHandle);
  if (NULL == pMappedRegion) {
    semGive(g_mappedListLock);
    return bFoundHandle ? ADMXRC2_INVALID_REGION : ADMXRC2_INVALID_HANDLE;
  }
  *ppVirtualBase = pMappedRegion->mapped.pVirtual;
  *pOffset = pMappedRegion->parameters.offset;
  *pLength = pMappedRegion->parameters.length;
  semGive(g_mappedListLock);
  
  return ADMXRC2_SUCCESS;
}
